﻿using System;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Runtime;

namespace WS2812_USB_UART
{
   
    internal enum IGJitter
    {
        High,
        Medium,
        Low,
    }
    

    internal class IntervalGenerator : IDisposable
    {
 
        #region Private Field

        private long _IntervalTime;
        private long _ElapstTicks;
        private IGJitter _IGJitter;

        private long ElapstTimeRest;
        private long LastTicks;

        private IntPtr ThreadHandle = IntPtr.Zero;

        private SpinWait spinWait = new SpinWait();

        #endregion


        #region Constructor

        public IntervalGenerator(long intervalTime, IGJitter iGJitter)
        {
            GCSettings.LatencyMode = GCLatencyMode.Batch;
            
            TimerResolution.CheckSetTimerResolution(TimerResolution.MaximumResolution);

            Thread.Yield();

            ThreadHandle = GetCurrentThread();

            IntervalTime = intervalTime;

            _IGJitter = iGJitter;

            ResetInterval();
        }

        #endregion


        #region Dispose

        private bool disposed;
        
        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing) { /*Managed*/ }

                if (ThreadHandle != IntPtr.Zero)
                {
                    CloseHandle(ThreadHandle);
                    ThreadHandle = IntPtr.Zero;
                }
                
                disposed = true;
            }
        }   // Dispose


        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }


        ~IntervalGenerator()
        {
            Dispose(false);
        }
        
        #endregion


        #region Public Methoden

        // Intervall Time lesen setzen
        public long IntervalTime
        {
            get { return _IntervalTime; }

            set
            {
                _IntervalTime = value;
                ResetInterval();
            }
        }
        

        // Elapset Time lesen
        public long ElapstTime
        {
            get
            {
                UpdateElapstTime();
                return ((_ElapstTicks * 1000000) / Stopwatch.Frequency);
            }
        }
        

        // Intervall rücksetzen
        public void ResetInterval()
        {
            LastTicks = Stopwatch.GetTimestamp();
            _ElapstTicks = 0;
            ElapstTimeRest = 0;
        }
        

        // Warten bis Intervall vorbei
        public int WaitForInterval()
        {
            SetThreadPriority(ThreadHandle, NativThreadPriority.THREAD_PRIORITY_HIGHEST);

            TimerResolution.CheckSetTimerResolution(TimerResolution.MaximumResolution);

            long cr = TimerResolution.CurrentResolution / 10;

            long msWait = (_IntervalTime - ElapstTime - cr) / 1000;

            if (msWait > int.MaxValue || msWait < 0)
            {
                ResetInterval();
                return -1;
            }

            if (msWait > 0)
                Thread.Sleep((int)msWait);

            if (_IGJitter == IGJitter.High)
            {
                if (ElapstTime < _IntervalTime)
                    while (ElapstTime < _IntervalTime) { spinWait.SpinOnce(); }
            }
            else
            {
                if (_IGJitter == IGJitter.Medium)
                    if (ElapstTime < _IntervalTime - cr)
                        while (ElapstTime < _IntervalTime - cr) { spinWait.SpinOnce(); }

                SetThreadPriority(ThreadHandle, NativThreadPriority.THREAD_PRIORITY_TIME_CRITICAL);

                while (ElapstTime < _IntervalTime) { Thread.SpinWait(10); }
            }

            int ret =  SubTimeSpan(_IntervalTime);
            

            return ret;
        }

        #endregion


        #region Private Methoden

        private void UpdateElapstTime()
        {
            long DeltaTicks;

            DeltaTicks = (Stopwatch.GetTimestamp() - LastTicks);
            LastTicks += DeltaTicks;

            _ElapstTicks += DeltaTicks;
        }
        

        private int SubTimeSpan(long SubTimeSpan)
        {
            if (SubTimeSpan == 0)
                return 0;

            int Count = 0;

            while (ElapstTime >= SubTimeSpan)
            {
                Count++;

                _ElapstTicks -= (SubTimeSpan * Stopwatch.Frequency) / 1000000;

                ElapstTimeRest += (SubTimeSpan * Stopwatch.Frequency) % 1000000;

                if (ElapstTimeRest >= 1000000)
                {
                    ElapstTimeRest -= 1000000;
                    _ElapstTicks -= 1;
                }
            }
            return Count;
        }

        #endregion


        #region Native Win32 API

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        internal static extern IntPtr GetCurrentThreadId();

        [DllImport("kernel32.dll")]
        static extern IntPtr GetCurrentThread();

        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        static extern bool CloseHandle(IntPtr hObject);

        [DllImport("kernel32.dll")]
        static extern NativThreadPriority GetThreadPriority(IntPtr hThread);

        [DllImport("kernel32.dll")]
        static extern bool SetThreadPriority(IntPtr hThread, NativThreadPriority nPriority);

        enum NativThreadPriority : int
        {
            THREAD_MODE_BACKGROUND_BEGIN = 0x00010000,
            THREAD_MODE_BACKGROUND_END = 0x00020000,
            THREAD_PRIORITY_ABOVE_NORMAL = 1,
            THREAD_PRIORITY_BELOW_NORMAL = -1,
            THREAD_PRIORITY_HIGHEST = 2,
            THREAD_PRIORITY_IDLE = -15,
            THREAD_PRIORITY_LOWEST = -2,
            THREAD_PRIORITY_NORMAL = 0,
            THREAD_PRIORITY_TIME_CRITICAL = 15
        }

        #endregion

    }
}
