﻿using System;
using System.IO.Ports;

// PL2303_Prolific_DriverInstaller_v1.5.0.exe für Win8

namespace WS2812_USB_UART
{

    internal enum USB_Device
    {
        LS_PL2303HXA,   // LowSpeed Profil 620 kBaud
        HS_PL2303HXA,   // HighSpeed Profil 780 kBaud
        LS_PL2303TA,    // LowSpeed Profil 570 kBaud
        //FT232RL,  // Geht nicht bei 3MBaud, Verschluckt sporadisch Daten (Win7, XP gleiches Verhalten)
        //CP2102,   // Geht nicht Stoppbit zu lang / Baudrate zu gering 2MBaut
    }

    
    internal class WS_UART
    {

        #region Puplic Fields

        public readonly USB_Device Device;
        public readonly byte[] LutRgbUart;
        public readonly int LedCount;
        public readonly byte[] RGB_Array;
        public readonly byte[] UartArray;

        #endregion
        

        #region Constructor

        public WS_UART(USB_Device device) : this(device, 0) { }
        public WS_UART(USB_Device device, int ledCount)
        {
            Device = device;

            LutRgbUart = new byte[LutRgbUartSize];

            InitializeLUT();

            
            if (ledCount == 0)
                return;

            byte[] rgbArray = null;
            byte[] uartArray = null;

            CreateArrays(ref rgbArray, ref uartArray, ledCount);

            LedCount = ledCount;
            RGB_Array = rgbArray;
            UartArray = uartArray;

            RgbArrayToWsUartArray();
        }

        #endregion


        #region LUT RGB UART

        public const int LutRgbUartSize = 256 * 4;

        //      S0123 456ss
        //00    10000 10000   0x10
        //01    10000 11100   0x70
        //10    11100 10000   0x13
        //11    11100 11100   0x73
        // LS_PL2023HXA
        // LS_PL2023TA
        private const byte P1_LutBit00 = 0x10 ^ 0xff;
        private const byte P1_LutBit01 = 0xe0 ^ 0xff;
        private const byte P1_LutBit10 = 0x13 ^ 0xff;
        private const byte P1_LutBit11 = 0xe3 ^ 0xff;
        private readonly byte[] P1_LutBit = { P1_LutBit00, P1_LutBit01, P1_LutBit10, P1_LutBit11 };

        //      S0123 456ss
        //00    11000 11000   0x31
        //01    11000 11100   0x71
        //10    11100 11000   0x33
        //11    11100 11100   0x73
        // HS_PL2023HXA
        private const byte P2_LutBit00 = 0x31 ^ 0xff;
        private const byte P2_LutBit01 = 0x71 ^ 0xff;
        private const byte P2_LutBit10 = 0x33 ^ 0xff;
        private const byte P2_LutBit11 = 0x73 ^ 0xff;
        private readonly byte[] P2_LutBit = { P2_LutBit00, P2_LutBit01, P2_LutBit10, P2_LutBit11 };

        private void InitializeLUT()
        {

            byte[] LutBit = null;

            switch (Device)
            {
                case USB_Device.LS_PL2303HXA:
                    LutBit = P1_LutBit;
                    break;
                case USB_Device.HS_PL2303HXA:
                    LutBit = P2_LutBit;
                    break;
                case USB_Device.LS_PL2303TA:
                    LutBit = P1_LutBit;
                    break;
            }
            
            int indexLut = 4;
            int j = 0;

            for (int i = 0; i < 256; i++)
            {
                for (j = 0; j < 8; j += 2)
                    LutRgbUart[--indexLut] = LutBit[(i >> j) & 0x03];
                indexLut += 8;
            }

        }

        #endregion


        #region Public Methods

        public static void CreateArrays(ref byte[] rgbArray, ref byte[] uartArray, int ledCount)
        {
            if (ledCount == 0)
                throw new Exception("ledCount == 0");

            rgbArray = new byte[ledCount * 3];
            uartArray = new byte[ledCount * 3 * 4];
        }


        public void SetupSerialPort(SerialPort serialPort)
        { SetupSerialPort(serialPort, Device, LedCount); }
        
        public static void SetupSerialPort(SerialPort serialPort, USB_Device dev, int ledCount)
        {

            serialPort.Parity = Parity.None;
            serialPort.StopBits = StopBits.One;
            serialPort.WriteBufferSize = ledCount * 3 * 4;
            
            switch (dev)
            {
                case USB_Device.LS_PL2303HXA:
                    serialPort.BaudRate = 3000000;       
                    serialPort.DataBits = 7;
                    break;

                case USB_Device.HS_PL2303HXA:
                    serialPort.BaudRate = 4500000;     
                    serialPort.DataBits = 7;
                    break;
                case USB_Device.LS_PL2303TA:
                    serialPort.BaudRate = 3000000;     
                    serialPort.DataBits = 7;
                    break;
            }
        }


        public void RgbArrayToWsUartArray()
        {
            RgbArrayToWsUartArray(RGB_Array, UartArray, LutRgbUart);
        }


        public static void RgbArrayToWsUartArray(byte[] rgbArray, byte[] uartArray, byte[] lutRgbUart)
        {
            if (rgbArray == null)
                throw new NullReferenceException("rgbArray");

            if (uartArray == null)
                throw new NullReferenceException("wsUartArray");

            if (lutRgbUart == null)
                throw new NullReferenceException("lutRgbUart");

            if (rgbArray.Length % 3 != 0)
                throw new Exception("rgbArray.Length is not valide");

            if (rgbArray.Length * 4 != uartArray.Length)
                throw new Exception("wsUartArray.Lengts is not valide");

            if (lutRgbUart.Length != LutRgbUartSize)
                throw new Exception("lutRgbUart.Length is not valide");


            int lengt = rgbArray.Length / 3;
            int iRgbA = 0;
            int iUartA = 0;
            int j, k1, k2, k3;

            for (int i = 0; i < lengt; i++)
            {
                k1 = rgbArray[iRgbA + 1] * 4;
                k2 = rgbArray[iRgbA + 0] * 4;
                k3 = rgbArray[iRgbA + 2] * 4;

                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k1 + j]; }
                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k2 + j]; }
                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k3 + j]; }
                iRgbA += 3;
            }
        }


        #endregion
        
    }
}
